Index: if_athvar.h
===================================================================
--- if_athvar.h	(revision 223525)
+++ if_athvar.h	(working copy)
@@ -254,7 +254,8 @@
 				sc_tdma	    : 1,/* TDMA in use */
 				sc_setcca   : 1,/* set/clr CCA with TDMA */
 				sc_resetcal : 1,/* reset cal state next trip */
-				sc_rxslink  : 1;/* do self-linked final descriptor */
+				sc_rxslink  : 1,/* do self-linked final descriptor */
+				sc_bexpect  : 1;/* expecting becaons to be RXed */
 	uint32_t		sc_eerd;	/* regdomain from EEPROM */
 	uint32_t		sc_eecc;	/* country code from EEPROM */
 						/* rate tables */
@@ -358,6 +359,8 @@
 	int			sc_txchainmask;	/* currently configured TX chainmask */
 	int			sc_rxchainmask;	/* currently configured RX chainmask */
 
+	struct callout		sc_bexpect_ch;	/* beacon expect callout */
+
 	/* DFS related state */
 	void			*sc_dfs;	/* Used by an optional DFS module */
 	int			sc_dodfs;	/* Whether to enable DFS rx filter bits */
Index: if_ath.c
===================================================================
--- if_ath.c	(revision 223568)
+++ if_ath.c	(working copy)
@@ -184,6 +184,8 @@
 static void	ath_scan_start(struct ieee80211com *);
 static void	ath_scan_end(struct ieee80211com *);
 static void	ath_set_channel(struct ieee80211com *);
+static void	ath_beacon_expect_reset(struct ath_softc *sc);
+static void	ath_beacon_expect_schedule(struct ath_softc *sc, int when);
 static void	ath_calibrate(void *);
 static int	ath_newstate(struct ieee80211vap *, enum ieee80211_state, int);
 static void	ath_setup_stationkey(struct ieee80211_node *);
@@ -387,6 +389,7 @@
 	}
 	callout_init_mtx(&sc->sc_cal_ch, &sc->sc_mtx, 0);
 	callout_init_mtx(&sc->sc_wd_ch, &sc->sc_mtx, 0);
+	callout_init_mtx(&sc->sc_bexpect_ch, &sc->sc_mtx, 0);
 
 	ATH_TXBUF_LOCK_INIT(sc);
 
@@ -2746,6 +2749,9 @@
 	u_int32_t nexttbtt, intval, tsftu;
 	u_int64_t tsf;
 
+	/* Clear any pending beacon expect timer */
+	ath_beacon_expect_reset(sc);
+
 	if (vap == NULL)
 		vap = TAILQ_FIRST(&ic->ic_vaps);	/* XXX */
 	ni = vap->iv_bss;
@@ -4666,6 +4672,58 @@
 }
 
 /*
+ * A beacon hasn't been heard in the given timeframe;
+ * write the last known beacon config to the hardware
+ * so beacon miss events begin occuring again.
+ */
+static void
+ath_beacon_expect(void *arg)
+{
+	struct ath_softc *sc = (struct ath_softc *) arg;
+	sc->sc_bexpect = 0;
+	DPRINTF(sc, ATH_DEBUG_BEACON, "%s: beacons not seen, kick!\n", __func__);
+	ath_beacon_config(sc, NULL);
+}
+
+/*
+ * Clear the beacon expect callout.
+ */
+static void
+ath_beacon_expect_reset(struct ath_softc *sc)
+{
+	if (! sc->sc_bexpect)
+		return;
+	sc->sc_bexpect = 0;
+	DPRINTF(sc, ATH_DEBUG_BEACON, "%s: clearing beacon expect\n", __func__);
+	callout_drain(&sc->sc_bexpect_ch);
+}
+
+/*
+ * Expect beacons in 'when' seconds.
+ *
+ * If they don't occur, begin the process of signalling
+ * the upper levels that beacons are being missed.
+ *
+ * This is done for now by simply writing the beacon config
+ * and letting the bmiss timer/interrupt handle things.
+ */
+static void
+ath_beacon_expect_schedule(struct ath_softc *sc, int when)
+{
+	sc->sc_bexpect = 1;
+	DPRINTF(sc, ATH_DEBUG_BEACON,
+	    "%s: scheduling beacon config in %d seconds\n",
+	    __func__, when);
+	callout_drain(&sc->sc_bexpect_ch);
+	/*
+	 * Add 1 second to 'when' to ensure the hostap actually
+	 * starts transmitting.
+	 */
+	callout_reset(&sc->sc_bexpect_ch, (when + 1) * hz,
+	    ath_beacon_expect, sc);
+}
+
+/*
  * Walk the vap list and check if there any vap's in RUN state.
  */
 static int
@@ -4748,6 +4806,28 @@
 	}
 
 	/*
+	 * If there's been a state transition from CSA -> RUN,
+	 * start a task to force reset the beacon timers
+	 * after CAC (default to 60 seconds.) Do this for non-
+	 * hostap modes.
+	 *
+	 * If we don't begin hearing beacons by then, we need
+	 * start registering beacon misses with net80211 or
+	 * rescanning won't occur and the interface will sit
+	 * here forever.
+	 *
+	 * XXX this requires that an association actually
+	 * XXX   occured sometime in the past!
+	 */
+	if (stamode && vap->iv_state == IEEE80211_S_CSA &&
+	    nstate == IEEE80211_S_RUN) {
+		DPRINTF(sc, ATH_DEBUG_STATE | ATH_DEBUG_BEACON,
+		    "%s: CSA->RUN detected; expecting beacons to appear"
+		    " in 60 seconds\n", __func__);
+		ath_beacon_expect_schedule(sc, 60);	/* XXX hard-coded CAC */
+	}
+
+	/*
 	 * Invoke the parent method to do net80211 work.
 	 */
 	error = avp->av_newstate(vap, nstate, arg);